/*---------------------------------------------------------------------------
	Module.cpp for LoPassDelay2.sem
	Version: 4.0, August 2004
	 2004, Guido Scognamiglio, All Rights Reserved
	http://www.soundfonts.it/vst - guido@gmlab.it

	Built with:
	SynthEdit SDK
	 2002, Jeff M Soft und Hardware GmbH, All Rights Reserved
---------------------------------------------------------------------------*/

#include "math.h"
#include "Module.h"
#include "SEMod_struct.h"
#include "SEPin.h"
#include "assert.h"

// define some constants to make referencing in/outs clearer
#define PN_INPUT_IN		0
#define PN_DTIME_IN		1
#define PN_FBACK_IN		2
#define PN_CUTOFF_IN	3
#define PN_RESONANCE_IN	4
#define PN_WETMIX_IN	5
#define PN_OVERDRIVE_IN	6
#define PN_DELAY_TYPE	7
#define PN_DELAY_MODE	8
#define PN_TIME_MODE	9
#define PN_AUDIO_OUT_L	10
#define PN_AUDIO_OUT_R	11

#define TYPE_NORMAL		0
#define TYPE_REVERSE	1

#define MODE_MONO		0
#define MODE_PINGPONG	1
#define MODE_PANNER		2

#define TIME_MODE_SECONDS	0
#define TIME_MODE_BPM		1

#define pi 3.141592f

// handy function to fix denormals.
static const float anti_denormal = 1e-18f;
inline void kill_denormal(float &p_sample ) { p_sample += anti_denormal;p_sample -= anti_denormal;};


// The most important part, processing the audio
// main process
void Module::sub_process(long buffer_offset, long sampleFrames) {

	// Retrieve Host Tempo
	VstTimeInfo *ti = 0;
	ti = (VstTimeInfo *) CallHost(seaudioMasterGetTime, 0, kVstTempoValid|kVstPpqPosValid, 0 );

	float *input = buffer_offset + input_in;
	float *dtime = buffer_offset + dtime_in;
	float *fback = buffer_offset + fback_in;
	float *cutoff = buffer_offset + cutoff_in;
	float *resonance = buffer_offset + resonance_in;
	float *wetmix = buffer_offset + wetmix_in;
	float *overdrive = buffer_offset + overdrive_in;
	float *out_l = buffer_offset + audio_out_l;
	float *out_r = buffer_offset + audio_out_r;

	// Keep d.time in range
	if (*dtime < 0.0001) *dtime = 0.0001;
	if (*dtime > 1) *dtime = 1;

	switch (time_mode) {
		case TIME_MODE_SECONDS:
			buffersize = (int) 10 * *dtime * samplerate;
			break;
		case TIME_MODE_BPM:
			int idx = 18 * *dtime;
			if (ti->tempo != 0) {
				buffersize = (60 * samplerate / ti->tempo ) * (4 / time_divisions[idx]);
			} else {
				buffersize = (int) 10 * *dtime * samplerate;
			}
			break;
	}


	// Read Cutoff PIN
	c0 = *cutoff;
	if (c0 > 0.99) c0 = 0.99;
	if (c0 < 0.01) c0 = 0.01;
	// Read Resonance PIN
	c1 = *resonance;
	if (c1 > 0.99) c1 = 0.99;
	if (c1 < 0.01) c1 = 0.01;

	// calculate filter temporary variables
	t0 = cos(pi * (0.99 * c0 + 0.01));
	t1 = sin(pi * (0.99 * c0 + 0.01)) / (3.5 * (c1 + 0.001));
	t2 = 1 + t1;
	// calculate filter coefficients
	a0 = (1 - t0) / (2 * t2);
	a1 = (1 - t0) / t2;
	a2 = (1 - t0) / (2 * t2);
	b1 = -2 * t0 / t2;
	b2 = ((1 - t1) / t2);

	while (--sampleFrames >= 0) { // main loop

		// Delay Line
		//this_sample = (*input * *wetmix) + (buffer[count] * *fback);
		this_sample = *wetmix * buffer[count]* *fback;
		float res = *input + this_sample;

		if (*overdrive > 0) { this_sample = (1.f - *overdrive) * tanh(this_sample * *overdrive * 10.f); }

		switch (d_mode) {
			case MODE_MONO:
				*out_l = *out_r = this_sample;
				break;
			case MODE_PINGPONG:
				/*
				if (count >= buffersize) {
					pan_position = prev_pan_position == 0.0f ? 1.0f : 0.0f;
				}
				prev_pan_position = pan_position;
				*/
				count2 += pan_direction;
				if (count2 <= 0) { pan_direction = 1; pan_position = 1.0f; }
				if (count2 >= buffersize) { pan_direction = -1; pan_position = 0.0f; }
				*out_l = this_sample * pan_position;
				*out_r = this_sample * (1.0f-pan_position);
				break;
			case MODE_PANNER:
				count2 += pan_direction;
				if (count2 <= 0) pan_direction = 1;
				if (count2 >= buffersize) pan_direction = -1;
				pan_position = float(count2) / float(buffersize) * 1.5f;
				*out_l = this_sample * pan_position;
				*out_r = this_sample * (1.5f-pan_position);
				break;
		}

		// apply filter to feedback chain
		yold2 = yold1;
		yold1 = y;
		y = a2 * xold2 + a1 * xold1 + a0 * res - b2 * yold2 - b1 * yold1;
		kill_denormal(y);
		xold2 = xold1;
		xold1 = res;

//		buffer[count] = y;	// Store next buffer
		buffer[count] = (this_sample + y) * 0.5f;	// Store next buffer (interpolated ??)

//		if (*overdrive > 0) { buffer[count] = tanh(buffer[count] * *overdrive * 10.f); }

		switch (d_type) {
			case TYPE_NORMAL:
				count++;
				count %= buffersize;	// if buffersize reaches end, start from 0
				break;
			case TYPE_REVERSE:
				count += delay_direction;
				if (count >= buffersize) delay_direction = -1;
				if (count <= 0) delay_direction = 1;
				break;
		}


		*input++;
		*out_l++;
		*out_r++;

		//*dtime++;
		//*fback++;
		//*cutoff++;

	}
}

/*
void Module::sub_process_static(long buffer_offset, long sampleFrames )
{
}
*/

void Module::sub_process_static(long buffer_offset, long sampleframes)
{
	sub_process( buffer_offset, sampleframes );

	static_count = static_count - sampleframes;
	if( static_count <= 0 )
	{
		CallHost(seaudioMasterSleepMode);
	}
}

// Class Constructor
Module::Module(seaudioMasterCallback2 seaudioMaster, void *p_resvd1) : SEModule_base(seaudioMaster, p_resvd1) , buffer(0)
{
}

// Class Destructor
Module::~Module()
{
	// This is where you free any memory/resources your module has created

	// free the memory used by the delay buffer
	delete [] buffer;
}

void Module::open() {
	// this macro switches the active processing function
	// e.g you may have several versions of your code, optimised for a certain situation
	SET_PROCESS_FUNC(Module::sub_process);
	SEModule_base::open(); // call the base class

	// Initialize Filter Variables
	c0=0.0;c1=0.0;
	t0=0.0;t1=0.0;t2=0.0;
	xold1=0.0;xold2=0.0;
	yold1=0.0;yold2=0.0;
	y=0;

	// Initialize Buffer
	samplerate = getSampleRate();
	buffersize = samplerate * 10;	// 10 second
	delete [] buffer;
	buffer = new FSAMPLE[buffersize];
	memset(buffer, 0, buffersize * sizeof(FSAMPLE) ); // clear buffer
	count = count2 = 0;
	pan_direction = delay_direction = 1;
	pan_position = 0.0f;
	prev_pan_position = 1.0f;


	// Time divisions for BPM mode (t = triplets, d = dotted)
	time_divisions[0] = 1;				// 1 bar
	time_divisions[1] = 2;				// half bar
	time_divisions[2] = 4;				// 1/4
	time_divisions[3] = 8;				// 1/8
	time_divisions[4] = 16;				// 1/16
	time_divisions[5] = 32;				// 1/32
	time_divisions[6] = 64;				// 1/64
	time_divisions[7] = 3;				// 1/2t
	time_divisions[8] = 6;				// 1/4t
	time_divisions[9] = 12;				// 1/8t
	time_divisions[10] = 24;			// 1/16t
	time_divisions[11] = 48;			// 1/32t
	time_divisions[12] = 96;			// 1/64t
	time_divisions[13] = 1.333;			// 1/2d
	time_divisions[14] = 2.666;			// 1/4d
	time_divisions[15] = 5.333;			// 1/8d
	time_divisions[16] = 10.666;		// 1/16d
	time_divisions[17] = 21.333;		// 1/32d
	time_divisions[18] = 42.666;		// 1/64d


	getPin(PN_AUDIO_OUT_L)->	TransmitStatusChange( SampleClock(), ST_RUN );
	getPin(PN_AUDIO_OUT_R)->	TransmitStatusChange( SampleClock(), ST_RUN );

}

// describe the module
bool Module::getModuleProperties (SEModuleProperties* properties) {
	properties->name = "LoPassDelay v2";	// this is the name the end-user will see

	// return a unique string.
	// if posible include manufacturer and plugin identity
	// this is used internally by SE to identify the plug. No two plugs may have the same id.
	properties->id = "LoPassDelay v2 by Guido Scognamiglio";

	// Info, may include Author, Web page whatever
	properties->about = "LoPassDelay v2 by Guido Scognamiglio\nhttp://www.soundfonts.it/vst\nziokiller@tiscali.it";
	return true;
}

// describe the pins (plugs)
bool Module::getPinProperties (long index, SEPinProperties* properties) {
	switch(index) {

	case PN_INPUT_IN:
		properties->name				= "Input";
		properties->variable_address	= &input_in;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;
	case PN_DTIME_IN:
		properties->name				= "D.Time";
		properties->variable_address	= &dtime_in;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "1";
		break;
	case PN_FBACK_IN:
		properties->name				= "F.Back";
		properties->variable_address	= &fback_in;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "5";
		break;
	case PN_CUTOFF_IN:
		properties->name				= "Cutoff";
		properties->variable_address	= &cutoff_in;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "10";
		break;
	case PN_RESONANCE_IN:
		properties->name				= "Resonance";
		properties->variable_address	= &resonance_in;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;
	case PN_WETMIX_IN:
		properties->name				= "Volume";
		properties->variable_address	= &wetmix_in;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "10";
		break;
	case PN_OVERDRIVE_IN:
		properties->name				= "Overdrive";
		properties->variable_address	= &overdrive_in;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;
	case PN_DELAY_TYPE:
		properties->name				= "Delay Type";
		properties->variable_address	= &d_type;
		properties->datatype_extra		= "Normal,Reverse";
		properties->default_value		= "0";
		properties->direction			= DR_IN;
		properties->datatype			= DT_ENUM;
		break;
	case PN_DELAY_MODE:
		properties->name				= "Delay Mode";
		properties->variable_address	= &d_mode;
		properties->datatype_extra		= "Mono,Ping-Pong,Panner";
		properties->default_value		= "0";
		properties->direction			= DR_IN;
		properties->datatype			= DT_ENUM;
		break;
	case PN_TIME_MODE:
		properties->name				= "Time Mode";
		properties->variable_address	= &time_mode;
		properties->datatype_extra		= "Seconds,BPM";
		properties->default_value		= "0";
		properties->direction			= DR_IN;
		properties->datatype			= DT_ENUM;
		break;
	case PN_AUDIO_OUT_L:
		properties->name				= "Out Left";
		properties->variable_address	= &audio_out_l;
		properties->direction			= DR_OUT;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;
	case PN_AUDIO_OUT_R:
		properties->name				= "Out Right";
		properties->variable_address	= &audio_out_r;
		properties->direction			= DR_OUT;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;
	default:
		return false; // host will ask for plugs 0,1,2,3 etc. return false to signal when done
	};
	return true;
}

// An input plug has changed value
void Module::OnPlugStateChange(SEPin *pin)
{
	// Syncronize the two incrementors

	if( pin->getPinID() == PN_DELAY_MODE ) {
		count2 = count;
		pan_position = 0.0f;
		prev_pan_position = 1.0f;
	}

}